# Standard modules
import logging
import subprocess
import re
import shlex
from argparse import Namespace
from pathlib import Path
from threading import Thread
from queue import Queue, Empty

# Pip modules
from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, TabbedContent, TabPane, Label, Input, Button, Collapsible, DirectoryTree, Log
from textual.binding import Binding
from textual.containers import Container, Horizontal

# Custom modules
from .ddrescue_tui_helper import LEGEND
from .ddrescue_tui_widgets import PixelWidget, StatusWidget


def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line)
    out.close()


class TUIddrescue(App):
    CSS_PATH = "ddrescue_tui.tcss"
    BINDINGS = [
        Binding(key="q", action="exit", description="Exit"),
        Binding(key="+", action="interval(2)", description="Change Interval", key_display="+/-"),
        Binding(key="-", action="interval(0.5)", show=False),
        Binding(key="d", action="toggle_dark", description="Toggle dark mode"),
        Binding(key="c", action="cancel", description="Cancel ddrescue", show=False)
    ]
    selected_input: Path | None = None
    selected_output: Path | None = None
    output_name: str | None = None
    custom_args: str | None = None
    ddrescue_process = None

    def __init__(self, args: Namespace):
        self.args = args
        super().__init__()

    def compose(self) -> ComposeResult:
        """Create child widgets for the application."""
        yield Header()
        with TabbedContent(initial="map"):
            with TabPane("Map", id="map"):
                yield PixelWidget(self.args)
                yield Label(LEGEND)
            with TabPane("ddrescue", id="output"):
                yield Container(
                    Horizontal(
                        Container(
                            Label("Select input:"),
                            DirectoryTree("/", id="input_path"),
                        ),
                        Container(
                            Label("Select output file or folder:"),
                            DirectoryTree("/", id="output_path")
                        )
                    ),
                    Collapsible(
                        Label("If an output name is specified, "
                              "the output file will be created in the selected directory or,"
                              " in case of a selected file, in the same directory. "
                              "A mapfile is always created as '<output name>.map'.", ),
                        title="Info",
                        collapsed_symbol="ⓘ >", expanded_symbol="ⓘ v"
                    ),
                    Input(placeholder="Output name (optional)", id="output_name"),
                    Input(placeholder="Custom arguments (optional)", id="custom_args"),
                    Button("Run ddrescue", id="submit"),
                    id="command_container"
                )
                yield Log(id="command_log", highlight=True)
        yield StatusWidget()
        yield Footer()

    def on_mount(self) -> None:
        """Titles for the header bar"""
        self.title = "ddrescue-tui"
        self.sub_title = "a pretty terminal user interface for ddrescue"
        if not self.args.sudo:
            self.query_one(TabbedContent).hide_tab("output")
        if self.args.action == "run":
            self.call_after_refresh(
                self.run_ddrescue,
                Path(self.args.input),
                Path(self.args.output),
                self.args.output_name,
                self.args.arguments
            )

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark

    def action_exit(self) -> None:
        """Exit the application."""
        self.exit()

    def action_interval(self, factor: float) -> None:
        """Increase the interval of the image reload"""
        pw = self.query_one(PixelWidget)
        pw.reload_timer(pw.interval * factor)
        logging.debug(f"interval: {pw.interval}")
        self.query_one(StatusWidget).interval = pw.interval

    def action_cancel(self) -> None:
        """Cancel the current ddrescue"""
        self.remove_class("running")
        self.bind("c", "cancel", description="Cancel ddrescue", show=False)
        self.query_one(Footer).refresh()

    def on_pixel_widget_info(self, msg: PixelWidget.Info) -> None:
        """Update the status bar information on "Info" message"""
        status = self.query_one(StatusWidget)
        status.version = msg.info.get("version") or ""
        status.action = msg.info.get("action") or "Running..."
        self.query_one(Log).clear()
        self.query_one(Log).write(msg.info.get("command"))

    def on_directory_tree_directory_selected(self, evt: DirectoryTree.DirectorySelected | DirectoryTree.FileSelected):
        """Tree item selected in command container"""
        match evt.control.id:
            case "input_path":
                self.selected_input = evt.path
            case "output_path":
                self.selected_output = evt.path

    def on_directory_tree_file_selected(self, evt: DirectoryTree.FileSelected):
        """Directory tree item selected in command container -> forward to general method"""
        self.on_directory_tree_directory_selected(evt)

    def on_input_changed(self, evt: Input.Changed) -> None:
        """Input changed in command_container for output_name or custom_args -> Update variable"""
        match evt.control.id:
            case "output_name":
                self.output_name = evt.value
            case "custom_args":
                self.custom_args = evt.value

    def on_button_pressed(self, evt: Button.Pressed) -> None:
        """Try to run ddrescue"""
        if not self.selected_input or not self.selected_output:
            # No input or output selected
            self.notify("Input and output are required", title="Could not run ddrescue", severity="warning")
            return

        self.run_ddrescue(self.selected_input, self.selected_output, self.output_name or '', self.custom_args or '')

    async def run_ddrescue(self, input_path: Path, output_path: Path, output_name: str, custom_args: str) -> None:
        """Run ddrescue either from command_container or directly from cli arguments"""
        # Combine arguments
        if output_name:
            if output_path.is_dir():
                output_path = output_path / output_name
            else:
                output_path = output_path.parent / output_name
        elif output_path.is_dir():
            output_path = output_path / f"{input_path.name}.img"
        map_path = output_path.parent / f"{output_path.stem}.map"
        cmd = f"ddrescue --mapfile-interval 1 {custom_args or ''} {input_path} {output_path} {map_path}"

        # UI changes
        self.notify(f"Command: {cmd}", title="Starting ddrescue", severity="information")
        self.add_class("running", update=True)
        self.bind("c", "cancel", description="Cancel ddrescue", show=True)
        self.query_one(Footer).refresh()
        log = self.query_one(Log)
        log.clear()
        log.write_line(f"$ {cmd}")
        pixel = self.query_one(PixelWidget)
        pixel.mapfile = map_path
        pixel.reload_timer(pixel.interval)

        # Run combined command
        # asyncio.create_task(self.start_ddrescue_async(shlex.split(cmd), log, map_path))
        proc = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=1, close_fds=True)
        q = Queue()
        t = Thread(target=enqueue_output, args=(proc.stdout, q), daemon=True)
        t.start()
        while proc.poll() is not None:
            try:
                if line := q.get_nowait():
                    # Count and remove ANSI line up codes
                    line_clean, up_count = re.subn(r"\x1b\[A", "", line.decode())
                    if up_count:
                        old_lines = list(log.lines[:-up_count])
                        old_lines.append(line_clean)
                        log.clear()
                        log.write_lines(old_lines)
                    else:
                        log.write_line(line_clean)
                else:
                    break
            except Empty:
                continue
