# Standard modules
import logging
from pathlib import Path
from argparse import Namespace

# Pip modules
from textual import events, widget, reactive, message
from textual.geometry import Size
from textual.widgets import Static
from rich_pixels import Pixels

# Custom modules
from .ddrescue_tui_parser import parse_mapfile
from .ddrescue_tui_plotter import heatmap_image


class StatusWidget(widget.Widget):
    """This widget shows the status bar at the bottom of the screen."""
    interval: reactive.reactive[float] = reactive.reactive(32)
    version: reactive.reactive[str] = reactive.reactive("")
    action: reactive.reactive[str] = reactive.reactive("")

    def render(self) -> str:
        return f"{self.version} | Reload interval: {self.interval}s | {self.action}"


class PixelWidget(Static):
    """This widget displays the parsed and plotted mapfile"""
    class Info(message.Message):
        """This message is sent with basic information for every mapfile update"""
        def __init__(self, info: dict) -> None:
            self.info = info
            super().__init__()

    class NoMapfile(message.Message):
        """This message is sent when no (valid) mapfile is set"""
        def __init__(self, last_mapfile: Path) -> None:
            self.last_mapfile = last_mapfile
            super().__init__()

    interval: float = None
    timer: events.Timer = None
    mapfile: Path | None
    width, height = 90, 30

    def __init__(self, args: Namespace) -> None:
        self.args = args
        super().__init__()

    def _on_mount(self, _) -> None:
        """Initial setup mapfile, image and timer"""
        self.mapfile = Path(self.args.mapfile) if hasattr(self.args, "mapfile") and self.args.mapfile else None
        self.loading = True
        self.load_image()
        # Wait a little bit until parent container is initialized
        self.set_timer(0.1, lambda: self.reload_timer(self.args.interval))

    def get_content_width(self, container: Size, viewport: Size) -> int:
        """Force custom size from the beginning"""
        return self.width

    def get_content_height(self, container: Size, viewport: Size, width: int) -> int:
        """Force custom size from the beginning"""
        return self.height

    def reload_timer(self, new_interval: float) -> None:
        """(Re-)Sets the timer to a new interval"""
        if self.timer:
            self.timer.stop()
        self.load_image()
        self.interval = new_interval
        self.timer = self.set_interval(self.interval, self.load_image)

    def load_image(self) -> None:
        """Tries to parse the mapfile to update the image"""
        # Get size from parent container if already initialized
        if self.container_size.width != 0 and self.container_size.height != 0:
            logging.debug(f"x {self.container_size.width}/y {self.container_size.height}")
            self.width = max(int(self.container_size.width / 2), 2)
            self.height = max(int(self.container_size.height), 2)

        # Parse mapfile
        data_arr, data_general = parse_mapfile(self.mapfile, self.width, self.height)

        # Mapfile invalid or inexistant
        if data_arr is None:
            old_mapfile = self.mapfile
            self.notify(f"Mapfile {self.mapfile} invalid")
            self.post_message(self.NoMapfile(old_mapfile))
            return

        # Create and display plot and send message about general data
        self.post_message(self.Info(data_general))
        image = heatmap_image(data_arr, self.width, self.height)
        self.update(Pixels.from_segments(Pixels._segments_from_image(image, (self.width, self.height))))
        self.loading = False
